﻿// COSC 4P76 - Assignment 2
// Michael Gircys (mg12vp, 5269550)
// 2016-03-17

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

namespace KMeansClusterer
{
    // Performs performance analysis on partition-based clustering results
    public class ClusterAnalytics
    {

        // Intra: within single cluster
        public enum IntraDistanceType
        {
            Max,
            Mean,
            DistFromMean
        }
        // Inter: between clusters
        public enum InterDistanceType
        {
            Closest,
            Furthest,
            CentroidDist
        }


        private DistanceFunctionDelegate  DistanceFunc;
        private DistanceFunctionType     _DistanceType;
        public  DistanceFunctionType      DistanceType
        {
            get { return _DistanceType; }
            set { _DistanceType = value; DistanceFunc = DistanceFunctions.Get(value); }
        }

        
        public ClusterAnalytics( DistanceFunctionType distMetric )
        {
            DistanceType = distMetric;
        }


        // Calculate a Dunn-Like index based on the type of intra- and inter-cluster distance measurements.
        public double Dunn_Index( DataTable dt,
            IntraDistanceType intraType = IntraDistanceType.Max, 
            InterDistanceType interType = InterDistanceType.Closest
        )
        {
            var k = dt.AsEnumerable().Max( p => (int)p["ClusterGroup"] ) + 1;

            var aggInter = new AggregateSet();
            var aggIntra = new AggregateSet();
            
            for(int k1 = 0; k1 < k; k1++)
                aggIntra.Add( IntraClusterDistance( dt, k1, intraType ) );
            for(int k1 = 0;    k1 < k-1; k1++)
            for(int k2 = k1+1; k2 < k-0; k2++)
                aggInter.Add( InterClusterDistance( dt, k1, k2, interType ) );

            var Dunn = Double.PositiveInfinity;
            if( aggIntra.Max != 0 )
                Dunn = aggInter.Min / aggIntra.Max;

            return Dunn;
        }

        // Calculate and Intra-Cluster distance (distance within a single cluster).
        public double IntraClusterDistance( DataTable dt, int Cluster, IntraDistanceType distMethod )
        {
            var dimens = dt.Columns.Count;

            // IF we don't have 2 points, then there's no distance between them.
            // This also helps us from screwing up the aggregate in the Dunn Index calculation.
            if(dt.Rows.Count < 2)
                return 0;

            // Get the points of the cluster.
            DataTable clusterPoints = dt;
            if(dt.Columns.Contains("ClusterGroup"))
            {
                var clusterPointsList = dt.AsEnumerable().Where( p => 
                        (int)p["ClusterGroup"]==Cluster && 
                        !(bool)p["IsCentroid"] 
                    );

                if(clusterPointsList.Count() < 2) 
                    return 0;

                clusterPoints = clusterPointsList.CopyToDataTable();
                dimens -= 2;
            }

            var distAgg = new AggregateSet();
            if(distMethod == IntraDistanceType.DistFromMean)
            {
                // Compare to mean.

                // Get mean point.
                var meanPoint = new double[dimens];
                var meanPointAgg = new AggregateSet[dimens];
                for( int d = 0; d < dimens; d++) 
                    meanPointAgg[d] = new AggregateSet();
                foreach( DataRow point in clusterPoints.Rows )
                for( int d = 0; d < dimens; d++)
                    meanPointAgg[d].Add( (double)point[d] );
                for( int d = 0; d < dimens; d++) 
                    meanPoint[d] = meanPointAgg[d].Avg;

                // Get distances from mean point
                foreach( DataRow point in clusterPoints.Rows )
                {
                    var dist = DistanceFunc.Invoke( point.ItemArray.Take(dimens).OfType<double>(), meanPoint );
                    distAgg.Add(dist);
                }
            }
            else
            {
                // Compare amongst one another.
                for( int i1 = 0;    i1 < clusterPoints.Rows.Count-1; i1++ )
                for( int i2 = i1+1; i2 < clusterPoints.Rows.Count-0; i2++ )
                {
                    var dist = DistanceFunc.Invoke( 
                            clusterPoints.Rows[i1].ItemArray.Take(dimens).OfType<double>(),
                            clusterPoints.Rows[i2].ItemArray.Take(dimens).OfType<double>()
                        );
                    distAgg.Add(dist);
                }
            }

            switch(distMethod)
            {
                case IntraDistanceType.Max:          return distAgg.Max;
                case IntraDistanceType.Mean:         return distAgg.Avg;
                case IntraDistanceType.DistFromMean: return distAgg.Avg;
                default: return 0.0;
            }
        }

        // Calculate an Inter-Cluster distance (distance between two clusters).
        public double InterClusterDistance( DataTable dt, int Cluster1, int Cluster2, InterDistanceType distMethod )
        {
            var dimens = dt.Columns.Count - 2;

            // Get the points of the clusters.
            var points1List = dt.AsEnumerable().Where( p => 
                    (int)p["ClusterGroup"] == Cluster1 && 
                    (bool)p["IsCentroid"] == (distMethod==InterDistanceType.CentroidDist)
                );
            var points2List = dt.AsEnumerable().Where( p => 
                    (int)p["ClusterGroup"] == Cluster2 && 
                    (bool)p["IsCentroid"] == (distMethod==InterDistanceType.CentroidDist)
                );

            // If a cluster is empty, return +Inf to not screw up the minimum (in the Dunn Index)
            if( points1List.Count() < 1 || points2List.Count() < 1 )
                return Double.PositiveInfinity;

            var points1 = points1List.CopyToDataTable();
            var points2 = points2List.CopyToDataTable();

            var distAgg = new AggregateSet();
            
            // Find the extrema of the distances between all points of C1 and all points of C2
            foreach( DataRow point1 in points1.Rows )
            foreach( DataRow point2 in points2.Rows )
            {
                var dist = DistanceFunc.Invoke(
                        point1.ItemArray.Take(dimens).OfType<double>(),
                        point2.ItemArray.Take(dimens).OfType<double>()
                    );
                distAgg.Add( dist );
            }

            switch(distMethod)
            {
                case InterDistanceType.Closest:      return distAgg.Min;
                case InterDistanceType.Furthest:     return distAgg.Max;
                case InterDistanceType.CentroidDist: return distAgg.Avg;
                default: return 0.0;
            }
        }


    }
}
